#include "riscos_module_header.h"
#include "utils.h"
#include "syscalls.h"

#define KEYPAD_ADDR 0x4a
#define KEYPAD_OFFSET 0xd2

#define KEYP_CTRL 0x00
#define KEYP_DEB 0x01
#define KEYP_LONG_KEY 0x02
#define KEYP_LK_PTV 0x03
#define KEYP_TIMEOUT_L 0x04
#define KEYP_TIMEOUT_H 0x05
#define KEYP_KBC 0x06
#define KEYP_KBR 0x07
#define KEYP_SMS 0x08
#define KEYP_FULL_CODE_7_0 0x09
#define KEYP_FULL_CODE_15_8 0x0a
#define KEYP_FULL_CODE_23_16 0x0b
#define KEYP_FULL_CODE_31_24 0x0c
#define KEYP_FULL_CODE_39_32 0x0d
#define KEYP_FULL_CODE_47_40 0x0e
#define KEYP_FULL_CODE_55_48 0x0f
#define KEYP_FULL_CODE_63_56 0x10
#define KEYP_ISR1 0x11
#define KEYP_IMR1 0x12
#define KEYP_ISR2 0x13
#define KEYP_IMR2 0x14
#define KEYP_SIR 0x15
#define KEYP_EDR 0x16
#define KEYP_SIH_CTRL 0x17

#define KEY(row, column, key) do { \
	table[((row) << 3) + (column)] = (key); \
	} while(0)

#define KEY_ESC 0
#define KEY_F1 1
#define KEY_F2 2
#define KEY_F3 3
#define KEY_F4 4
#define KEY_F5 5
#define KEY_F6 6
#define KEY_F7 7
#define KEY_F8 8
#define KEY_F9 9
#define KEY_F10 10
#define KEY_F11 11
#define KEY_F12 12
#define KEY_PRINT 13
#define KEY_SCROLL_LOCK 14
#define KEY_BREAK 15
#define KEY_TILDE 16
#define KEY_1 17
#define KEY_2 18
#define KEY_3 19
#define KEY_4 20
#define KEY_5 21
#define KEY_6 22
#define KEY_7 23
#define KEY_8 24
#define KEY_9 25
#define KEY_0 26
#define KEY_MINUS 27
#define KEY_EQUAL 28
#define KEY_F18 29
#define KEY_BACKSPACE 30
#define KEY_INSERT 31
#define KEY_HOME 32
#define KEY_PAGEUP 33
#define KEY_NUMLOCK 34
#define KEY_F16 37
#define KEY_TAB 38
#define KEY_Q 39
#define KEY_W 40
#define KEY_E 41
#define KEY_R 42
#define KEY_T 43
#define KEY_Y 44
#define KEY_U 45
#define KEY_I 46
#define KEY_O 47
#define KEY_P 48
#define KEY_LEFTBRACE 49
#define KEY_RIGHTBRACE 50
#define KEY_BACKSLASH 51
#define KEY_DELETE 52
#define KEY_COPY 53
#define KEY_PAGEDOWN 54
#define KEY_KPMINUS 58
#define KEY_LEFTCTRL 59
#define KEY_A 60
#define KEY_S 61
#define KEY_D 62
#define KEY_F 63
#define KEY_G 64
#define KEY_H 65
#define KEY_J 66
#define KEY_K 67
#define KEY_L 68
#define KEY_SEMICOLON 69
#define KEY_APOSTROPHE 70
#define KEY_RETURN 71
#define KEY_KPPLUS 75
#define KEY_LEFTSHIFT 76
#define KEY_Z 78
#define KEY_X 79
#define KEY_C 80
#define KEY_V 81
#define KEY_B 82
#define KEY_N 83
#define KEY_M 84
#define KEY_COMMA 85
#define KEY_DOT 86
#define KEY_SLASH 87
#define KEY_R_SHIFT 88
#define KEY_UP 89
#define KEY_CAPSLOCK 93
#define KEY_LEFTALT 94
#define KEY_SPACE 95
#define KEY_REIGHTALT 96
#define KEY_RIGHTCTRL 97
#define KEY_LEFT 98
#define KEY_DOWN 99
#define KEY_RIGHT 100
#define KEY_ENTER 103
#define KEY_SELECT 112
#define KEY_MENU 113
#define KEY_ADJUST 114

typedef struct pndkbd_data_s {
	unsigned int *gpio_base;
	unsigned int counter;
	unsigned int status;
	unsigned int plattform;
	char temp[32];
	unsigned char keyp[8];
	unsigned int errors;
	/** True, if keyb_update() is or will running. */
	int updating;
	int keytable[64];
	int fnkeytable[64];
} pndkbd_data_t;


void keyb_handler(void);

void prints(const char *text)
{
	OS_Write0(relocate((char *) text));
}


void print_hex(unsigned int value)
{
	char buffer[9];

	buffer[0] = 0;
	OS_ConvertHex8(value, buffer, sizeof(buffer));
	OS_Write0(buffer);
}

unsigned int strnlen(const char *text, unsigned int len)
{
	unsigned int i;

	for (i = 0; i < len; i++) {
		if (text[i] == 0) {
			return i;
		}
	}
	return len;
}

void print_dec4(unsigned int value, unsigned int n)
{
	unsigned int i;
	char buffer[9];
	unsigned int len;

	buffer[0] = 0;
	OS_ConvertInteger3(value, buffer, sizeof(buffer));
	len = strnlen(buffer, sizeof(buffer));
	for (i = len; i < n; i++) {
		prints(" ");
	}
	OS_Write0(buffer);
}

void my_memset(void *p, int c, int n)
{
	int i;
	char *b;

	for (i = 0, b = p; i < n; i++) {
		b[i] = c;
	}
}

unsigned char i2c_read(pndkbd_data_t *env, unsigned char addr, unsigned char reg)
{
	iic_transfer_t t[2];
	unsigned char rv = 0;

	my_memset(t, 0, sizeof(t));
	t[0].addr = (addr << 1) | I2C_WRITE;
	t[0].d.data = &reg;
	t[0].len = sizeof(reg);

	t[1].addr = (addr << 1) | I2C_READ;
	t[1].d.data = &rv;
	t[1].len = sizeof(rv);

	if (OS_IICOp(t, sizeof(t) / sizeof(t[0]))) {
		env->errors++;
	}

	return rv;
}

void i2c_write(pndkbd_data_t *env, unsigned char addr, unsigned char reg, unsigned char value)
{
	iic_transfer_t t[1];
	unsigned char data[2];

	data[0] = reg;
	data[1] = value;
	my_memset(t, 0, sizeof(t));
	t[0].addr = (addr << 1) | I2C_WRITE;
	t[0].d.data = &data;
	t[0].len = sizeof(data);

	if (OS_IICOp(t, sizeof(t) / sizeof(t[0]))) {
		env->errors++;
	}
}

void dump_regs(pndkbd_data_t *env)
{
#if 0
	int i;

	for (i = KEYP_FULL_CODE_7_0; i <= KEYP_FULL_CODE_63_56; i++) {
		unsigned char v;

		v = i2c_read(env, KEYPAD_ADDR, KEYPAD_OFFSET + i);
		prints("I2C Reg &");
		print_hex(i);
		prints(": &");
		print_hex(v);
		OS_NewLine();
	}
#endif
	prints("I2C Errors: &");
	print_hex(env->errors);
	OS_NewLine();
#if 0
	prints("Keyp 0: &");
	print_hex(env->keyp[0]);
	OS_NewLine();
	prints("Keyp 1: &");
	print_hex(env->keyp[1]);
	OS_NewLine();
#endif
}

void writereg(pndkbd_data_t *env, unsigned char reg, unsigned char value)
{
	i2c_write(env, KEYPAD_ADDR, KEYPAD_OFFSET + reg, value);
}

void keypad_init(pndkbd_data_t *env)
{
	int *table;
	unsigned int i;

#if 0
	prints("Module started");
	OS_NewLine();
	prints("GPIO Base Address &");
	print_hex((unsigned int) env->gpio_base);
	OS_NewLine();
#endif

	env->updating = 0;
	env->errors = 0;
	env->plattform = OS_ReadSysInfo(8);
	for (i = 0; i < sizeof(env->keyp)/sizeof(env->keyp[0]); i++) {
		env->keyp[i] = 0;
	}
	/* Enable controller, with hardware decoding but not autorepeat */
	writereg(env, KEYP_CTRL, 0x4b);
	/* Enable TO rising and KP rising and falling edge detection */
	writereg(env, KEYP_EDR, 0x23);
	/* Set PTV prescaler Field */
	writereg(env, KEYP_LK_PTV, 0x80);
	/* Set key debounce time to 20 ms */
	writereg(env, KEYP_DEB, 0x13);
	/* Set timeout period to 100 ms */
	writereg(env, KEYP_TIMEOUT_L, 0xc8);
	writereg(env, KEYP_TIMEOUT_H, 0xc8);
	/*
	 * Enable Clear-on-Read; disable remembering events that fire
	 * after the IRQ but before our handler acks (reads) them.
	 */
	writereg(env, KEYP_SIH_CTRL, 0x06);

	/* Initialize table for mapping keypad to RISC OS low-level key number. */
	table = env->keytable;
	my_memset(table, -1, sizeof(env->keytable));

	KEY(0, 0, KEY_9);
	KEY(0, 1, KEY_8);
	KEY(0, 2, KEY_I);
	KEY(0, 3, KEY_J);
	KEY(0, 4, KEY_N);
	KEY(0, 5, KEY_M);
	KEY(1, 0, KEY_0);
	KEY(1, 1, KEY_7);
	KEY(1, 2, KEY_U);
	KEY(1, 3, KEY_H);
	KEY(1, 4, KEY_B);
	KEY(1, 5, KEY_SPACE);
	KEY(2, 0, KEY_BACKSPACE);
	KEY(2, 1, KEY_6);
	KEY(2, 2, KEY_Y);
	KEY(2, 3, KEY_G);
	KEY(2, 4, KEY_V);
	// TBD KEY(2, 5, KEY_FN);
	KEY(3, 0, KEY_O);
	KEY(3, 1, KEY_5);
	KEY(3, 2, KEY_T);
	KEY(3, 3, KEY_F);
	KEY(3, 4, KEY_C);
	KEY(4, 0, KEY_P);
	KEY(4, 1, KEY_4);
	KEY(4, 2, KEY_R);
	KEY(4, 3, KEY_D);
	KEY(4, 4, KEY_X);
	KEY(5, 0, KEY_K);
	KEY(5, 1, KEY_3);
	KEY(5, 2, KEY_E);
	KEY(5, 3, KEY_S);
	KEY(5, 4, KEY_Z);
	KEY(6, 0, KEY_L);
	KEY(6, 1, KEY_2);
	KEY(6, 2, KEY_W);
	KEY(6, 3, KEY_A);
	KEY(6, 4, KEY_DOT);
	KEY(7, 0, KEY_ENTER);
	KEY(7, 1, KEY_1);
	KEY(7, 2, KEY_Q);
	KEY(7, 3, KEY_LEFTSHIFT);
	KEY(7, 4, KEY_COMMA);

	/*
	 * Initialize table for mapping keypad to RISC OS low-level key number
	 * when FN is pressed
	 */
	table = env->fnkeytable;
	my_memset(table, -1, sizeof(env->fnkeytable));
	KEY(0, 0, KEY_F9);
	KEY(0, 1, KEY_F8);
	// TBD KEY(0, 2, KEY_BRIGHTNESSUP);
	// TBD KEY(0, 3, KEY_F13);		/* apostrophe, differs from Fn-A? */
	// TBD KEY(0, 4, KEY_F22);
	// TBD KEY(0, 5, KEY_F23);
	KEY(1, 0, KEY_F10);
	KEY(1, 1, KEY_F7);
	// TBD KEY(1, 2, KEY_BRIGHTNESSDOWN);
	// TBD KEY(1, 3, KEY_GRAVE);
	// TBD KEY(1, 4, KEY_F14);		/* pipe/bar */
	KEY(1, 5, KEY_TAB);
	KEY(2, 0, KEY_INSERT);
	KEY(2, 1, KEY_F6);
	// TBD KEY(2, 2, KEY_F15);		/* dash */
	KEY(2, 3, KEY_EQUAL);
	KEY(2, 4, KEY_F16);		/* # (pound/hash) */
	// TBD KEY(2, 5, KEY_FN);
	KEY(3, 0, KEY_F11);
	KEY(3, 1, KEY_F5);
	// TBD KEY(3, 2, KEY_F17);		/* ! */
	KEY(3, 3, KEY_KPPLUS);
	KEY(3, 4, KEY_BACKSLASH);
	KEY(4, 0, KEY_F12);
	KEY(4, 1, KEY_F4);
	KEY(4, 2, KEY_RIGHTBRACE);
	KEY(4, 3, KEY_KPMINUS);
	// TBD KEY(4, 4, KEY_QUESTION);
	KEY(5, 0, KEY_F18);		/* £ (pound) */
	KEY(5, 1, KEY_F3);
	KEY(5, 2, KEY_LEFTBRACE);
	// TBD KEY(5, 3, KEY_F19);		/* " */
	KEY(5, 4, KEY_SLASH);
	// TBD KEY(6, 0, KEY_YEN);
	KEY(6, 1, KEY_F2);
	// TBD KEY(6, 2, KEY_F20);		/* @ */
	KEY(6, 3, KEY_APOSTROPHE);
	// TBD KEY(6, 4, KEY_F21);		/* : */
	KEY(7, 0, KEY_ENTER);
	KEY(7, 1, KEY_F1);
	KEY(7, 2, KEY_ESC);
	KEY(7, 3, KEY_CAPSLOCK);
	KEY(7, 4, KEY_SEMICOLON);
}

void cmd_pndkbd(pndkbd_data_t *env)
{
	prints(" ____                 _                   ____      _     _      _   _");
	OS_NewLine();
	prints("|  _ \\ __ _ _ __   __| | ___  _ __ __ _  |  _ \\ ___| |__ (_)_ __| |_| |__");
	OS_NewLine();
	prints("| |_) / _` | '_ \\ / _` |/ _ \\| '__/ _` | | |_) / _ \\ '_ \\| | '__| __| '_ \\");
	OS_NewLine();
	prints("|  __/ (_| | | | | (_| | (_) | | | (_| | |  _ <  __/ |_) | | |  | |_| | | |");
	OS_NewLine();
	prints("|_|   \\__,_|_| |_|\\__,_|\\___/|_|  \\__,_| |_| \\_\\___|_.__/|_|_|   \\__|_| |_|");
	OS_NewLine();
	prints("                            ____                           _   _ _   _");
	OS_NewLine();
	prints("         _,-'-,_           / ___|___  _ __ ___  _ __   ___| |_(_) |_(_) ___  _ __");
	OS_NewLine();
	prints("     _,-'  _,_  '-,_      | |   / _ \\| '_ ` _ \\| '_ \\ / _ \\ __| | __| |/ _ \\| '_ \\");
	OS_NewLine();
	prints(" _,-'  _,-'   '-,_  '-,_  | |__| (_) | | | | | | |_) |  __/ |_| | |_| | (_) | | | |");
	OS_NewLine();
	prints("'-,_  '-,_     _,-'  _,-'  \\____\\___/|_| |_| |_| .__/ \\___|\\__|_|\\__|_|\\___/|_| |_|");
	OS_NewLine();
	prints("|-, '-,_  '| |'  _,-' ,-|                      |_|");
	OS_NewLine();
	prints("|  |    |  | |  |    |  |");
	OS_NewLine();
	prints("|  |    |  | |  |    |  |");
	OS_NewLine();
	prints("|  '-,_ |  | |  | _,-'  |");
	OS_NewLine();
	prints("'-,_   '|  | |  |'   _,-'");
	OS_NewLine();
	prints("    '-,_   | |   _,-'");
	OS_NewLine();
	prints("        '-,| |,-'");
	OS_NewLine();
	OS_NewLine();
	prints("This module adds support for the Open Pandora Keyboard");
	OS_NewLine();
	prints("Pandora Keyboard Status");
	OS_NewLine();

	prints("GPIO: &");
	print_hex(env->gpio_base[0x38 / 4]);
	OS_NewLine();

	prints("Status: &");
	print_hex(env->status);
	OS_NewLine();

	prints("Counter: &");
	print_hex(env->counter);
	OS_NewLine();

	prints("Plattform class: &");
	print_hex(env->plattform);
	OS_NewLine();

	dump_regs(env);
}

int cmd_pndkey(pndkbd_data_t *env, unsigned int type)
{
	switch (type) {
		case 0:
			prints("Please type *help pndkey");
			OS_NewLine();
			break;

		case 1: {
			unsigned int i;
			unsigned int len;

			prints("Key  Low  Key  Low  Key  Low  Key  Low");
			OS_NewLine();
			prints("--------------------------------------");
			OS_NewLine();
			len = (sizeof(env->keytable)/sizeof(env->keytable[0]))/2;
			for (i = 0; i < len; i++) {
				print_dec4(i, 3);
				prints(" ");
				print_dec4(env->keytable[i], 4);
				prints("  ");

				print_dec4(i + len, 3);
				prints(" ");
				print_dec4(env->keytable[i + len], 4);
				prints("  ");

				print_dec4(i + 64, 3);
				prints(" ");
				print_dec4(env->fnkeytable[i], 4);
				prints("  ");

				print_dec4(i + 64 + len, 3);
				prints(" ");
				print_dec4(env->fnkeytable[i + len], 4);
				OS_NewLine();
			}
			break;
		}

		default: {
			const char *text = (void *) type;
			unsigned int key;

#if 0
			prints("Set Open Pandora Key param: ");
			OS_Write0(text);
			OS_NewLine();
#endif

			key = OS_ReadUnsigned(10, text, 0);
			while(*text != 0) {
				if (*text == ' ') {
					text++;
					break;
				}
				text++;
			}
			if (key < 64) {
				env->keytable[key] = OS_ReadUnsigned(10, text, 0);
			} else if (key < 128) {
				env->fnkeytable[key - 64] = OS_ReadUnsigned(10, text, 0);
			} else {
				prints("Key is out of range");
				OS_NewLine();
			}
			break;
		}
	}
	return 0;
}

void keyb_trigger_update(pndkbd_data_t *env)
{
	if (!env->updating) {
		env->updating = 1;
		OS_AddCallBack(relocate(&keyb_handler), env);
	}
}

void keyb_update(pndkbd_data_t *env)
{
	unsigned int row;

	for (row = 0; row < sizeof(env->keyp)/sizeof(env->keyp[0]); row++)
	{
		int i;
		unsigned char changed;
		unsigned char state;

		/* Get state of current row. */
		state = i2c_read(env, KEYPAD_ADDR, KEYPAD_OFFSET + KEYP_FULL_CODE_7_0 + row);
		changed = env->keyp[row] ^ state;
		env->keyp[row] = state;

		for (i = 0; i < 8; i++) {
			unsigned char mask;
			mask = 1 << i;
			if (changed & mask) {
				int *table;
				int s;
				int key;

				if (env->keyp[2] & (1 << 5)) {
					/* FN is pressed. */
					table = env->fnkeytable;
				} else {
					table = env->keytable;
				}
				/* Convert to RISC OS low-level key number. */
				key = table[(row << 3) + i];
				if (key >= 0) {
					/* Key state has been changed. */
					if (state & mask) {
						/* Key pressed. */
						s = 2;
					} else {
						/* Key released. */
						s = 1;
					}
					OS_KeyV(s, key);
				}
			}
		}
	}
	env->updating = 0;
}
