@ RISC OS constants
XOS_WriteS = 0x20001
XOS_Write0 = 0x20002
XOS_NewLine = 0x20003
XOS_CLI = 0x20005
XOS_FSControl = 0x20029
XOS_ValidateAddress = 0x2003a
XMessageTrans_ErrorLookup = 0x61506
XOS_Module = 0x2001e
XOS_Claim = 0x2001f
XOS_Release = 0x20020
XOS_CallAVector = 0x20034
XOS_Memory = 0x20068
XOS_ConvertHex8 = 0x200d4
XOS_CallEvery = 0x2003c
XOS_RemoveTickerEvent = 0x2003d
XOS_AddCallBack = 0x20054
XOS_RemoveCallBack = 0x2005f

WORKSPACE_SIZE = 1024

GPIO_BASE = 0
COUNTER = GPIO_BASE + 4
STATUS = COUNTER + 4
PLATTFORM = STATUS + 4
TEMP = PLATTFORM + 4

GPIOIN = 0x38

	@ Help and Command keyword table
.global commands
commands:
	.string "pndkbd"
	.align
	.int command_pndkbd
	.int 0x00000000
	.int 0
	.int command_pndkbd_help

	.string "pndkey"
	.align
	.int command_pndkey
	.int 0x40020001
	.int 0
	.int command_pndkey_help

	.byte 0			@ Table terminator

command_pndkbd_help:
	.string "*pndkbd Show Open Pandora keyboard status"
command_pndkey_help:
	.string "*status pndkey\r\
Show low-level key number conversion table for\r\
Open Pandora Keyboard.\r\
\r\
*configure pndkey [key] [low-level key number]\r\
Set low-level key number for Open Pandora key.\r\
The normal keys have a range 0-63.\r\
The range 64-128 is for keys when FN is pressed\r\
at the same time.\r"
	.align

keytable:
	/* LEFT */
	.int 0x00000001
	.int 98
	/* RIGHT */
	.int 0x00000004
	.int 100
	/* Pandora Key -> Break */
	.int 0x00000008
	.int 15
	/* ALT */
	.int 0x00000010
	.int 94
	/* HOME */
	.int 0x00000020
	.int 32
	/* RIGHT SHIFT */
	.int 0x00000040
	.int 88
	/* DOWN */
	.int 0x00000080
	.int 99
	/* CTRL */
	.int 0x00000100
	.int 59
	/* RIGHT CTRL */
	.int 0x00000200
	.int 97
	/* PAGE DOWN */
	.int 0x00000400
	.int 54
	/* PAGE UP */
	.int 0x00002000
	.int 33
	/* UP */
	.int 0x00004000
	.int 89
	/* END */
	.int 0x00008000
	.int 53
keytable_end:

/* Entry:
 *   r10 = pointer to environment string
 *   r11 = I/O base or instantiation number
 *   r12 = pointer to private word for this instantiation
 *   r13 = stack pointer (supervisor)
 * Exit:
 *   r7-r11, r13 preserved
 *   other may be corrupted
 */
.global init
init:
	STMFD R13!,{R7-R11,LR}

	TEQ R12, #0
	BEQ reinit

	@ Claim some workspace
	MOV	R0, #6
	MOV	R3, #WORKSPACE_SIZE
	SWI	XOS_Module
	BVS init_return
	STR	R2, [R12]				@ Store workspace pointer in private word

	LDR	R12, [R12]				@ Get workspace pointer
reinit:

	MOV R1, #0
	STR R1, [R12, #COUNTER]

	MOV R1, #0xFF
	ORR R1, R1, #0xFF00
	STR R1, [R12, #STATUS]

	MOV R0, #1					@ Setup timer in centiseconds
	ADR R1, ticker_handler
	MOV R2, R12
	SWI XOS_CallEvery
	BVS init_return

	/* Only working on Pandora RISC OS 5. */
	@ Map IO memory for GPIO
	MOV R0, #13
	LDR R1, =0x49054000
	MOV R2, #0x200
	SWI XOS_Memory
	BVS init_return
	STR R3, [R12, #GPIO_BASE]

	MOV R0, #0					@ Install keyboard
	MOV R1, #2					@ PC/AT keyboard
	MOV R2, #0					@ unused before RISC OS 5
	MOV R9, #0x13				@ KeyV
	SWI XOS_CallAVector
	BVS init_return

	MOV R0, R12					@ First parameter is workspace
	BL keypad_init

init_return:
	LDMFD R13!,{R7-R11,PC}

/* Entry:
 *   r12 = pointer to private word for this instantiation
 *   r13 = stack pointer (supervisor)
 * Exit:
 *   r7-r11, r13 preserved
 *   other may be corrupted
 */
.global final
final:
	STMFD R13!,{R7-R11,LR}
	LDR	R12, [R12]				@ Get workspace pointer

	@ TBD: unpress all keys...

	@ Remove ticker handler
	ADR R0, ticker_handler
	MOV R1, R12
	SWI XOS_RemoveTickerEvent
	BVS final_return

	@ Remove callback
	ADR R0, keyb_handler
	MOV R1, R12
	SWI XOS_RemoveCallBack

	@ Free workspace
	MOV	r0, #7
	MOV	r2, r12
	SWI	XOS_Module
final_return:
	LDMFD R13!,{R7-R11,PC}		@ restore registers and exit

/* Entry:
 *   r12 = pointer to private word for this instantiation
 *   r13 = stack pointer (supervisor)
 * Exit:
 *   r7-r11, r13 preserved
 *   other may be corrupted
 */
command_pndkbd:
	STMFD R13!,{R7-R11,LR}
	LDR R0, [R12]
	BL cmd_pndkbd

	LDMFD R13!,{R7-R11,PC}		@ restore registers and exit

/* Entry:
 *   r12 = pointer to private word for this instantiation
 *   r13 = stack pointer (supervisor)
 * Exit:
 *   r7-r11, r13 preserved
 *   other may be corrupted
 */
command_pndkey:
	STMFD R13!,{R7-R11,LR}
	MOV R1, R0
	LDR R0, [R12]
	BL cmd_pndkey

	LDMFD R13!,{R7-R11,PC}		@ restore registers and exit

/* Entry:
 *   r12 = workspace pointer
 *   r13 = stack pointer (supervisor)
 * Exit:
 *   r0-r13 preserved
 *   other may be corrupted
 */
ticker_handler:
	STMFD R13!,{R0-R12,LR}

	LDR R1, [R12, #COUNTER]
	ADD	R1, R1, #1
	STR R1, [R12, #COUNTER]

	LDR R1, [R12, #STATUS]		@ Get previous status
	LDR R3, [R12, #GPIO_BASE]	@ Get GPIO base address
	LDR R4, [R3, #GPIOIN]		@ Read GPIO state

	CMN R4, #1
	BEQ gpio_fail				@ When module is tested in rpcemu, the GPIOIN value is 0xFFFFFFFF

	EOR R4, R4, #0x10			@ ALT key is inverted
	STR R4, [R12, #STATUS]		@ Save current status
	EOR R2, R1, R4				@ Check which bits are changed

	ADR R5, keytable

nextkey:
	LDR R6, [R5]				@ Get GPIO mask for button
	TST R2, R6
	BEQ inckey					@ Branch if bit hasn't been changed
	TST R4, R6
	MOVEQ R0, #2				@ key pressed
	MOVNE R0, #1				@ key released

	STMFD R13!,{R2-R12}			@ save register for SWI call
	LDR R1, [R5, #4]			@ Low level internal key number
	MOV R9, #0x13				@ KeyV
	SWI XOS_CallAVector
	LDMFD R13!,{R2-R12}			@ restore registers
inckey:
	ADD R5, R5, #8				@ next entry in keytable
	ADR R6, keytable_end
	TEQ R5, R6					@ check if keytable_end reached
	BNE nextkey

gpio_fail:
	MOV R0, R12					@ Pointer to workspace
	BL keyb_trigger_update		@ Call handler for keypad (I2C).

ticker_exit:
	LDMFD R13!,{R0-R12,LR}		@ restore registers and exit
	MOV PC, LR

/* Entry:
 *   r12 = workspace pointer
 *   r13 = stack pointer (supervisor)
 * Exit:
 *   r0-r13 preserved
 *   other may be corrupted
 */
.global keyb_handler
keyb_handler:
	STMFD R13!,{R0-R12,LR}

	MOV R0, R12					@ Pointer to workspace
	BL keyb_update				@ Call handler for keypad (I2C).

keyb_handler_exit:
	LDMFD R13!,{R0-R12,LR}		@ restore registers and exit
	MOV PC, LR
